/*
** Module   : ButtonGuiImpl.java
** Abstract : GUI push button implementation.
**
** Copyright (c) 2014-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 EVL 20141120 Initial version.
** 002 CA  20141205 Avoid unnecessary conversion, as the physical location/dimension can be 
**                  determined from existing APIs.  Do not draw directly from APIs which are   
**                  called from mouse events: instead, rely on repainting, which posts proper event
**                  to the event chain and process the drawing in the parent's context.
** 003 CA  20141206 Improvement of H002 - refresh needs to call repaint() always.
** 004 EVL 20141211 Detaching tooltip code from button widget.  Adding image loading for button.
** 005 EVL 20141226 Using images to draw button implementation.
** 006 EVL 20150115 Rework for image usage implementation.
** 007 EVL 20150121 Fixup image coordinates while loading.  Adding handling of the deferred
**                  images.
*/

package com.goldencode.p2j.ui.client.gui;

import java.awt.Color;
import java.awt.Font;
import java.awt.BasicStroke;
import java.awt.Stroke;
import java.awt.event.MouseEvent;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.GuiFontResolver.FontCache;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Specific GUI implementation of RECTANGLE widget.
 */
@SuppressWarnings("unchecked")
public class ButtonGuiImpl
extends Button<GuiOutputManager>
{
   /** Screen driver */
   private GuiDriver gd;

   /** Output manager */
   private GuiOutputManager screen;
   
   /** Coordinate converter */
   private CoordinatesConversion cc;

   /** Color resolver */
   private GuiColorResolver gc;
   
   /** Font resolver */
   private GuiFontResolver gf;
      
   /** Light shadow box color. */
   private Color lightShadow;

   /** Dark shadow box color. */
   private Color darkShadow;
   
   /** Very dark shadow box color. */
   private Color veryDarkShadow;
   
   /** Dotted line pattern for focused button text frame. */
   private Stroke dotStroke;
   
   /** Dotted line specific parameter. */
   private static final float DOT_STEP = 2.0f;
   
   /** Dotted line width. */
   private static final float DOT_LINE_WIDTH = 1.0f;
   
   /** Pressed button state flag. */
   private boolean pressed = false;
   
   /** Shift in XY to draw images on the button. */
   private static final int IMAGE_SHIFT = 4;
   
   /** Image to draw on button in up state. */
   private ImageGuiImpl imgUp = null; 
   
   /** Image to draw on button in pressed state. */
   private ImageGuiImpl imgDown = null; 
   
   /** Image to draw on button in disabled state. */
   private ImageGuiImpl imgDisabled = null; 
   
   /**
    * This constructor creates a button with a given text label.
    * 
    * @param   id
    *          The widget's ID.
    * @param   owner
    *          The parent container, frame or dialog which will include this
    *          button.
    * @param   text
    *          The string to be used as the button text.
    */
   ButtonGuiImpl(WidgetId id, Container<GuiOutputManager> owner, String text) 
   {
      super(id, owner, text);
   }

   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public ButtonGuiImpl(WidgetId id, WidgetConfig cfg)
   {
      super(id, (ButtonConfig) cfg);
   }
   
   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    wcfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, WidgetConfig wcfg)
   {
      super.initialize(id, wcfg);
      // screen driver
      gd = (GuiDriver) OutputManager.getDriver();
      // output manager
      screen = screen();
      cc = screen.coordinates();
      // font resolver
      gf = new GuiFontResolver(Window.resolveWindow(this), gd, config());
      // initialize the shadow colors
      lightShadow    = gd.getSysColor(SysColor.COLOR_WINDOW);
      darkShadow     = gd.getSysColor(SysColor.COLOR_3DSHADOW);
      veryDarkShadow = gd.getSysColor(SysColor.COLOR_3DDKSHADOW);
      // initialize dot pattern 
      dotStroke = new BasicStroke(DOT_LINE_WIDTH, BasicStroke.CAP_BUTT,
                                  BasicStroke.JOIN_BEVEL, 0.0f, // this must be just valid
                                  new float[] {DOT_STEP, DOT_STEP}, DOT_STEP);

      // register tooltip action listener
      gd.registerTooltipWorker(this, gf.font().font, ToolTip.isActive(config().tooltip));
      // handle pending images to preload
      loadPendingImages();
   }
   
   /**
    * Sets the new image for given button widget on the client side.
    * 
    * @param    type
    *           Image type to load.  Up, Down or Insensitive for buttons.
    * @param    image
    *           A new image array data.
    *
    * @return   <code>TRUE</code> if operation successful <code>FALSE</code> otherwise.
    */
   public boolean setImage(int type, byte[] image)
   {
      boolean result = false;
      
      switch (type)
      {
         case ButtonConfig.IMAGE_UP:
         {
            if (imgUp == null)
            {
               imgUp = new ImageGuiImpl(WidgetId.nextID());
            }
            result = imgUp.setImage(image);

            // fixup button coordinates
            if (result)
            {
               fixupImageCoord(imgUp);
            }
            break;
         }
         case ButtonConfig.IMAGE_DOWN:
         {
            if (imgDown == null)
            {
               imgDown = new ImageGuiImpl(WidgetId.nextID());
            }
            result = imgDown.setImage(image);

            // fixup button coordinates
            if (result)
            {
               fixupImageCoord(imgDown);
            }
            break;
         }
         case ButtonConfig.IMAGE_INS:
         {
            if (imgDisabled == null)
            {
               imgDisabled = new ImageGuiImpl(WidgetId.nextID());
            }
            result = imgDisabled.setImage(image);

            // fixup button coordinates
            if (result)
            {
               fixupImageCoord(imgDisabled);
            }
            break;
         }
      }
      
      return result;
   }
   
   /**
    * Loads the new image for given button widget from the client side filesystem.
    * 
    * @param    type
    *           Image type to load.  Up, Down or Insensitive for buttons.
    * @param    fullName
    *           The fully qualified file name for the image to be loaded.
    *
    * @return   <code>TRUE</code> if operation successful <code>FALSE</code> otherwise.
    */
   public boolean loadImage(int type, String fullName)
   {
      boolean result = false;
      
      switch (type)
      {
         case ButtonConfig.IMAGE_UP:
         {
            if (imgUp == null)
            {
               imgUp = new ImageGuiImpl(WidgetId.nextID());
            }
            result = imgUp.loadImage(fullName);
            
            // fixup button coordinates
            if (result)
            {
               fixupImageCoord(imgUp);
            }
            break;
         }
         case ButtonConfig.IMAGE_DOWN:
         {
            if (imgDown == null)
            {
               imgDown = new ImageGuiImpl(WidgetId.nextID());
            }
            result = imgDown.loadImage(fullName); 

            // fixup button coordinates
            if (result)
            {
               fixupImageCoord(imgDown);
            }
            break;
         }
         case ButtonConfig.IMAGE_INS:
         {
            if (imgDisabled == null)
            {
               imgDisabled = new ImageGuiImpl(WidgetId.nextID());
            }
            result = imgDisabled.loadImage(fullName); 

            // fixup button coordinates
            if (result)
            {
               fixupImageCoord(imgDisabled);
            }
            break;
         }
      }
      
      return result;
   }
   
   /**
    * In GUI mode, the drawing needs to be delayed.
    */
   @Override
   protected void refresh()
   {
      repaint();
   }
   
   /**
    * Draw the button. Called by this Button's parent container.
    */
   @Override
   public void draw()
   {
      // button options need to be refreshed before every painting
      final ButtonConfig config = (ButtonConfig) config();
      gc = new GuiColorResolver(gd, config, SysColor.COLOR_MENU, SysColor.COLOR_BTNTEXT);
      
      // determine colors to use
      gc.colors();
      
      // resolve font
      final String text = textLabel();
      final FontCache fntCache = gf.font();

      // TODO: find out what will be the right coordinates to use here
      final NativePoint p = physicalLocation();
      final NativeDimension d = physicalDimension();

      final NativeRectangle clip = clipRectangle(p, d);
      
      // TODO: clipping does not work properly, need to fix clipRectangle() in AbstractWidget 
      gd.draw(p, null, new Runnable()
      {

         @Override
         public void run()
         {
            // inside this call widget origin is always (1,1)
            // TODO: this is caused by the frame's 1px border... need to fix at the widget 
            // placement, not here. Check other cases too!
            NativePoint po = new NativePoint(1, 1);

            // button state flags
            boolean highlighted = highlighted();
            
            // focused button needs special highlight
            if (highlighted)
            {
               drawHighlight(po, d, gc.fgColor);
            }

            // draw button itself
            if (hasImages())
            {
               // button border is included in method below 
               drawButtonImages(po, d, highlighted, pressed);
            }
            else
            {
               // draw button border
               drawButtonBorder(po, d, highlighted, pressed);
               // abd the body
               drawButtonInternal(po, d, text, fntCache.font, highlighted, pressed);
            }
         }
         
      });
   }

   /**
    * Sets up initial image location inside the button widget.
    * 
    * @param   imgToFix
    *          Image object to fix the coordinates.
    */
   private void fixupImageCoord(ImageGuiImpl imgToFix)
   {
      imgToFix.setPhysicalLocation(IMAGE_SHIFT, IMAGE_SHIFT);
   }
   
   /**
    * Draw highlight frame indicating focused button.
    * 
    * @param   orig
    *          Rectangle origin.
    * @param   dim
    *          Rectangle size.
    * @param   color
    *          Border color.
    */
   private void drawHighlight(NativePoint orig, NativeDimension dim, Color color)
   {
      gd.setColor(color);
      gd.drawRect(orig.x, orig.y, dim.width, dim.height);
   }
   
   /**
    * Draw bottom border simulating 3D effect. 
    * 
    * @param   orig
    *          Rectangle origin.
    * @param   dim
    *          Rectangle size.
    * @param   focused
    *          Is the button focused.
    * @param   pressed
    *          Is the button pressed.
    */
   private void drawButtonBorder(NativePoint orig,
                                 NativeDimension dim,
                                 boolean focused,
                                 boolean pressed)
   {
      // coordinates are dependent from button state
      int x1 = focused ? orig.x + 1 : orig.x;
      int x2 = focused ? orig.x + dim.width - 1 : orig.x + dim.width;
      int y1 = focused ? orig.y + 1 : orig.y;
      int y2 = focused ? orig.y + dim.height - 1 : orig.y + dim.height;

      if (pressed)
      {
         gd.setColor(darkShadow);
         gd.drawRect(x1, y1, x2 - x1, y2 - y1);
      }
      else
      {
         gd.setColor(lightShadow);
         gd.drawLine(x1, y1, x1, y2 - 1);
         gd.drawLine(x1, y1, x2 - 1, y1);

         gd.setColor(darkShadow);
         gd.drawLine(x1 + 1, y2 - 1, x2 - 1, y2 - 1);
         gd.drawLine(x2 - 1, y2 - 1, x2 - 1, y1 + 1);

         gd.setColor(veryDarkShadow);
         gd.drawLine(x1, y2, x2, y2);
         gd.drawLine(x2, y2, x2, y1);
      }
   }
   
   /**
    * Draw button internals with different states.
    * 
    * @param   orig
    *          Button origin.
    * @param   dim
    *          Button size.
    * @param   label
    *          Button label.
    * @param   fnt
    *          Font to use for button label.
    * @param   focused
    *          Is the button focused.
    * @param   pressed
    *          Is the button pressed.
    */
   private void drawButtonInternal(NativePoint orig,
                                   NativeDimension dim,
                                   String label,
                                   Font fnt,
                                   boolean focused,
                                   boolean pressed)
   {
      // coordinates are dependent from button state
      int x = focused ? orig.x + 1 : orig.x;
      int y = focused ? orig.y + 1 : orig.y;
      int w = focused ? dim.width - 2 : dim.width;
      int h = focused ? dim.height - 2 : dim.height;

      // draw button body
      gd.setColor(gc.bgColor);
      if (pressed)
      {
         gd.fillRect(x + 1, y + 1, w - 1, h - 1);
      }
      else
      {
         gd.fillRect(x + 1, y + 1, w - 2, h - 2);
      }
      
      // draw button text
      gd.setGuiFont(fnt);
      int xt = orig.x - 1 + (dim.width - gd.getTextWidth(label, fnt))/2;
      int yt = orig.y - 1 + (dim.height + fnt.getSize())/2;
      gd.setFontStyle(fnt.getStyle());
      // enabled state controls text drawing
      if (isEnabled())
      {
         gd.setColor(gc.fgColor);
         if (pressed)
         {
            gd.drawString(label, xt + 1, yt + 1);
         }
         else
         {
            gd.drawString(label, xt, yt);
         }
      }
      else
      {
         gd.setColor(lightShadow);
         gd.drawString(label, xt + 1, yt + 1);
         gd.setColor(darkShadow);
         gd.drawString(label, xt, yt);
      }

      // draw dotted rectangle around focused button text
      if (isEnabled() && focused)
      {
         // set dotted line style
         gd.setLineStroke(dotStroke);
         int dx = (dim.width - gd.getTextWidth(label, fnt))/4;
         int dy = (dim.height - fnt.getSize())/4;
         if (pressed)
         {
            gd.drawRect(x + dx + 2, y + dy + 2, w - 2*dx - 2, h - 2*dy - 2);
         }
         else
         {
            gd.drawRect(x + dx + 1, y + dy + 1, w - 2*dx - 2, h - 2*dy - 2);
         }
         // restore default line style
         gd.setLineStroke(new BasicStroke());
      }
   }
   
   /**
    * Draw button internals with different states using preloaded images.
    * 
    * @param   orig
    *          Button origin.
    * @param   dim
    *          Button size.
    * @param   focused
    *          Is the button focused.
    * @param   pressed
    *          Is the button pressed.
    */
   private void drawButtonImages(NativePoint orig,
                                 NativeDimension dim,
                                 boolean focused,
                                 boolean pressed)
   {
      // coordinates are dependent from button state
      int x1 = orig.x + 1;
      int x2 = orig.x + dim.width - 1;
      int y1 = orig.y + 1;
      int y2 = orig.y + dim.height - 1;

      // draw 3D shadow for image button
      if (pressed)
      {
         gd.setColor(lightShadow);
         gd.drawLine(x1 + 1, y2, x2, y2);
         gd.drawLine(x2, y2, x2, y1);

         gd.setColor(darkShadow);
         gd.drawLine(x1 + 1, y2 - 1, x1 + 1, y1 + 1);
         gd.drawLine(x1 + 1, y1 + 1, x2 - 1, y1 + 1);

         gd.setColor(veryDarkShadow);
         gd.drawLine(x1, y1, x1, y2);
         gd.drawLine(x1, y1, x2, y1);
      }
      else
      {
         gd.setColor(lightShadow);
         gd.drawLine(x1, y1, x1, y2 - 1);
         gd.drawLine(x1, y1, x2 - 1, y1);

         gd.setColor(darkShadow);
         gd.drawLine(x1 + 1, y2 - 1, x2 - 1, y2 - 1);
         gd.drawLine(x2 - 1, y2 - 1, x2 - 1, y1 + 1);

         gd.setColor(veryDarkShadow);
         gd.drawLine(x1, y2, x2, y2);
         gd.drawLine(x2, y2, x2, y1);
      }
      // draw the light frame around image
      gd.setColor(gc.bgColor);
      if (pressed)
      {
         gd.drawRect(x1 + 2, y1 + 2, x2 - x1 - 3, y2 - y1 - 3);
      }
      else
      {
         gd.drawRect(x1 + 1, y1 + 1, x2 - x1 - 3, y2 - y1 - 3);
      }
      // draw button body images
      if (pressed)
      {
         // determine what image to draw
         ImageGuiImpl imgPressed = null;
         if (imgDown != null)
         {
            imgPressed = imgDown; 
         }
         else if (imgUp != null)
         {
            imgPressed = imgUp; 
         }
         // draw pressed state
         if (imgPressed != null)
         {
            // for pressed state need 1 pixel right-down shift 
            NativePoint location = imgPressed.physicalLocation();
            imgPressed.setPhysicalLocation(location.x + 1, location.y + 1);
            imgPressed.draw();
            // and restore origin after draw
            imgPressed.setPhysicalLocation(location.x, location.y);
         }
      }
      else
      {
         // determine what image to draw
         ImageGuiImpl imgReleased = null;
         if (!isEnabled() && imgDisabled != null)
         {
            imgReleased = imgDisabled; 
         }
         else if (imgUp != null)
         {
            imgReleased = imgUp; 
         }
         // draw released state
         if (imgReleased != null)
         {
            imgReleased.draw();
         }
      }
   }

   /**
    * Check if the button has any images to be used to draw button.
    * 
    * @return  <code>true</code> there is at least one image to use.
    */
   private boolean hasImages()
   {
      return imgUp != null || imgDown != null || imgDisabled != null;
   }

   /**
    * Refresh the image objects to be used to draw button in different states with images that was
    * saved before in pending storage area.
    */
   private void loadPendingImages()
   {
      String imgName = null;
      // down image first
      imgName = config().imageDown;
      if (imgName != null && imgName.length() > 0 && imgDown == null)
      {
         loadPendingImage(ButtonConfig.IMAGE_DOWN);
      }
      // up image next
      imgName = (config().imageUp != null) ? config().imageUp : config().image;
      if (imgName != null && imgName.length() > 0)
      {
         loadPendingImage(ButtonConfig.IMAGE_UP);
      }
      // image for disabled state
      imgName = config().imageIns;
      if (imgName != null && imgName.length() > 0)
      {
         loadPendingImage(ButtonConfig.IMAGE_INS);
      }
   }

   /**
    * Refresh the image object to be used to draw button in different states with images that was
    * saved before in pending storage area.
    * 
    * @param   type
    *          Image type to check.  Up, Down or Insensitive.
    */
   private void loadPendingImage(int type)
   {
      int thisId = config().id.getId();

      // first try to search for images loaded from application jar
      byte[] imgData = ImageHelper.getPendingImageData(thisId, type);
      if (imgData != null)
      {
         setImage(type, imgData);
      }
      else
      {
         String loadName = ImageHelper.getPendingImageName(thisId, type);
         if (loadName != null && loadName.length() > 0)
         {
            loadImage(type, loadName);
         }
      }
   }
   
   /**
    * Determine the physical bounds, relative to screen, expressed in units native to UI type.
    * 
    * @return  See above.
    */
   @Override
   public NativeRectangle physicalBounds()
   {
      // TODO: this is not correct just to verify mouse handling methods 
      return new NativeRectangle(cc.pointToNative(location()),
                                 cc.dimensionToNative(dimension()));
   }
   
   /**
    * Get widget height.
    * 
    * @return  widget height.
    */
   @Override
   public double height()
   {
      // have no images use regular calculation
      if (!hasImages())
      {
         return super.height();
      }
      
      // compute existed images heights
      double hUp   = imgUp != null ? imgUp.height() : 0.0;
      double hDown = imgDown != null ? imgDown.height() : 0.0;
      double hDis  = imgDisabled != null ? imgDisabled.height() : 0.0;
      
      // and return max value of them plus image insets
      return Math.max(hUp, Math.max(hDown, hDis)) +
             screen().coordinates().heightFromNative(2*IMAGE_SHIFT - 2);
      
   }

   /**
    * Get widget width.
    * 
    * @return  widget width.
    */
   @Override
   public double width()
   {
      // have no images use regular calculation
      if (!hasImages())
      {
         return super.width();
      }
      
      // compute existed images widths
      double wUp   = imgUp != null ? imgUp.width() : 0.0;
      double wDown = imgDown != null ? imgDown.width() : 0.0;
      double wDis  = imgDisabled != null ? imgDisabled.width() : 0.0;
      
      // and return max value of them plus image insets
      return Math.max(wUp, Math.max(wDown, wDis)) +
             screen().coordinates().widthFromNative(2*IMAGE_SHIFT - 2);
   }

   /**
    * Notification of a mouse clicked event occurred for this widget.  Simulate ENTER key typing.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseClicked(MouseEvent e)
   {
      // simulate enter key
      if (isEnabled())
      {
         EventManager.postEvent(new KeyInput(this, EventType.KEY_TYPED, Key.VK_ENTER));
      }
   }

   /**
    * Notification of a mouse pressed event occurred for this widget.  Redraw button in pressed
    * state and request focus.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mousePressed(MouseEvent e)
   {
      // push button effect
      if (isEnabled() && !pressed)
      {
         requestFocus();
         pressed = true;
         repaint();
      }
   }

   /**
    * Notification of a mouse released event occurred for this widget.  Redraw button in released
    * state.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseReleased(MouseEvent e)
   {
      // release button effect
      if (isEnabled() && pressed)
      {
         pressed = false;
         repaint();
      }
   }

   /**
    * Process basic events - TAB, BACK-TAB and ENTER key event.
    * 
    * @param   event
    *          Key event to process.
    */
   @Override
   protected void handleBasicEvents(KeyInput event)
   {
      // special processing for spacebar key in GUI mode
      int id = event.id();
      boolean needRefresh = false;
      
      if (isEnabled())
      {
         // any typed key will upress the button
         if (pressed && id == EventType.KEY_PRESSED)
         {
            pressed = false;
            // state changed, need repainting
            needRefresh = true;
         }
         else
         {
            // space bar is VK_ENTER with KEY_TYPED id
            if (event.keyCode() == Key.VK_ENTER && id == EventType.KEY_TYPED)
            {
               pressed = true;
               // state changed, need repainting
               needRefresh = true;
               // simulate enter key
               EventManager.postEvent(new ActionEvent(this, getCommand()));
            }
         }
         // redraw the button state
         if (needRefresh)
         {
            repaint();
         }
      }
      
      // all other should be processed as usual
      super.handleBasicEvents(event);
   }

   /**
    * Handle cursor movement keys.
    * 
    * @param   key
    *          Key code.
    * @param   frame
    *          Parent frame.
    * 
    * @return  <code>true</code> if any cursor key was found.
    */
   protected boolean handleCursorKeys(int key, Frame<GuiOutputManager> frame)
   {
      // cursor movement does not work as focus change for GUI button
      if (key != Keyboard.KA_CURSOR_LEFT &&
          key != Keyboard.KA_CURSOR_UP &&
          key != Keyboard.KA_CURSOR_RIGHT &&
          key != Keyboard.KA_CURSOR_DOWN)
      {
         return super.handleCursorKeys(key, frame);
      }
      else
      {
         return false;
      }
   }
}
